// parser2.c/cpp - easier to use, adds stack and xt dictionary
// The funcs not prefixed by 'skip', 'copy', or 'is' or 'put' 
// may need a 'obj_' prefix.  Here's hoping we can use the 
// simple names.

#define PARSER2_NO_WARN
#include "parser2.h"
#include "tmpstr.h"

#include <stdio.h>
#include <stdlib.h> // exit()
#include <malloc.h>
#include <string.h>
#include <stdarg.h> // va_list...

// a convenience macro, should also be used in parser files.
#define obj (*g_parser_obj)

// These it's relatively safe ;-) to use sprintf(obj.op...) 
// to set output text, but to be real safe 'obj_needs(0)' should
// be called afterward to assure the buffer has an opportunity
// to grow if necessary.
#define OBUF_INIT_SIZE 4096
#define OBUF_PADDING 4096 

// required global object
PARSER_OBJ* g_parser_obj;

////////////////////////////////////////////////////
// built-in functions

static char errormsg[1024];

void pvt_parser_init() {
  // main watch points
  obj.state = true;
  obj.op_watch = obj.outbuf;
  obj.op_ctr = 0;
  obj.nest_level = 0;
  
  obj.errorflag = false;
  obj.errormsg = errormsg;
  memset(errormsg, 0, sizeof(errormsg));
  obj.ip = obj.inbuf;
  obj.op = obj.outbuf;
  *obj.op = 0;
  memset(obj.pstack, 0, obj.pstack_size * sizeof(char*));
  obj.sp = 0;
  memset(obj.rstack, 0, obj.rstack_size * sizeof(xt_type));
  obj.rp = 0;
  memset(obj.dict, 0, obj.dict_size * sizeof(long));
  obj.dp = 0;
}

// Returns a new parser obj pointer
PARSER_OBJ* parser_new(
  const char* inbuffer, 
  int inlen, 
  bool dynalloc
)
{
  PARSER_OBJ* prev = g_parser_obj;
  g_parser_obj = (PARSER_OBJ*)malloc(sizeof(PARSER_OBJ));
  obj.prev = prev;
  obj.inbuflen = inlen;
  obj.inbuf = inbuffer;
  obj.iblimit = obj.inbuf + obj.inbuflen;
  obj.dynalloc = dynalloc;
  
  obj.outbuflen = OBUF_INIT_SIZE;
  obj.outbuf = (char*)malloc(OBUF_INIT_SIZE);
  obj.oblimit = obj.outbuf + OBUF_INIT_SIZE - OBUF_PADDING;
  
  obj.pstack_size = PARSER_STACK_SIZE; // by char** fields
  obj.pstack = (char**)malloc(obj.pstack_size * sizeof(char**));
  obj.dict_size = PARSER_DICT_SIZE; // by dword fields
  obj.dict = (xt_type*)malloc(obj.dict_size * sizeof(xt_type));
  obj.rstack_size = RETURN_STACK_SIZE; // by char** fields
  obj.rstack = (xt_type*)malloc(obj.rstack_size * sizeof(xt_type*));
  pvt_parser_init(); // 
  return g_parser_obj;
}

// frees memory including This, sets the global 
// object to prev-ious and also returns it.
PARSER_OBJ* parser_delete(PARSER_OBJ* This)
{
  PARSER_OBJ* prev = This->prev;
  if(This->outbuf) free(This->outbuf);
  if(This->dynalloc)
    free((void*)This->inbuf);
  free(This->pstack);
  free(This->rstack);
  free(This->dict);
  free(This);
  return g_parser_obj = prev;
}

// Inputs set to null will be ignored.
void obj_get_ipline(int* pline_no, int* pc_offs, char* linebuf, int bufmax)
{
  const char* lp = obj.inbuf;
  int c_offs = 0;
  int line_no = 1;
  int i;
  
  for(const char* p = obj.inbuf; p < obj.iblimit; p++)
  {
    if(p == obj.ip)
      break;
    
    if(*p == '\n')
    {
      if((p + 1) == obj.ip)
        break;      
      lp = p;
      c_offs = 0;
      line_no++;
    }
    else
      c_offs++;
  }
  
  if(line_no != 1)
  {
  
    // if offs still 0, try to back up to something non-zero
    if(c_offs == 0)
    {
      while((*lp == '\n') && (lp > obj.inbuf))
      {
        lp--;
        line_no--;
      }
      while((*lp != '\n') && (lp > obj.inbuf))
      {
        lp--;
        c_offs++;
      }
      lp++;
    }
    else if(*lp == '\n')
    {
      lp++;
      c_offs --;
    }
  }    
  
  if(pline_no)
    *pline_no = line_no;
  
  if(pc_offs)
    *pc_offs = c_offs;
  
  if(!linebuf || !bufmax)
    return;
  
  bufmax--;
  
  for(i = 0; i < bufmax; i++, lp++)
  {
    if(*lp == '\n')
      break;
    linebuf[i] = *lp;
  }
  linebuf[i] = 0;
}

// numBytes may be 0, for example to force the buffer to
// resize with at least OBUF_PADDING bytes.
bool obj_needs(int numBytes)
{
  while((numBytes + obj.op) > obj.oblimit)
  {
    int newsize = obj.outbuflen * 2;
    int op_offset = obj_get_op();
    obj.outbuf = (char*)realloc(obj.outbuf, newsize);
    obj.outbuflen = newsize;
    obj.oblimit = obj.outbuf + newsize - OBUF_PADDING;
    obj.op = obj.outbuf + op_offset;
  }
  return obj.state = true;
}

bool put(const char* text)
{
  int slen = strlen(text);
  obj_needs(slen + 1);
  strcpy(obj.op, text);
  obj.op += slen;
  return obj.state = true;
}

// putfmt is like sprintf(...). It puts formatted text
// at the output buffer and advances op.
bool putfmt(const char* fmt, ...)
{
  obj_needs(256); // make sure padding is plenty
  va_list params;
  va_start(params, fmt);
  obj.op += vsnprintf(obj.op, OBUF_PADDING, fmt, params);
  va_end(params);
  obj_needs(0);   // make sure padding is still reasonable
  return obj.state = true;
}

// sets errormsg and errorflag, returns true but doesn't
// change obj.state
bool obj_set_errormsg(const char* fmt, ...)
{
  va_list params;
  va_start(params, fmt);
  vsnprintf(obj.errormsg, sizeof(errormsg), fmt, params);
  va_end(params);
  obj.errorflag = true;
  return true;
}


bool is_eoi() {
  return obj.state = (obj.ip >= obj.iblimit) ? true : false;
}

bool not_eoi() { 
  return obj.state = !is_eoi();
}

PPTRS_OBJ* pvt_parser_save_ptrs(PPTRS_OBJ* p) {
  p->ip = obj.ip;
  p->svop = obj_get_op();
  p->sp = obj.sp;
  p->rp = obj.rp;
  p->dp = obj.dp;
  return p;
}

// returns true so it can be inlined but it does not change
// the parser's internal state var.  See OPT_RESTORE_PTRS()
// in the header file.
bool pvt_parser_restore_ptrs(PPTRS_OBJ* p) {
  obj.ip = p->ip;
  obj_set_op(p->svop);
  obj.sp = p->sp;
  obj.rp = p->rp;
  obj.dp = p->dp;  
  return true;
}

// true if string at ip = s
bool is(const char* s) {
  obj.state = true;
  int slen = strlen(s);
  if(!is_nchars(slen)) // check ip is long enough to be it
    return obj.state = false;

  if(memcmp((void*)obj.ip, (void*)s, slen) != 0)
    obj.state = false;
  return obj.state;
}

// copy if ip = s
bool copy(const char* s) {
  if(!is(s))
    return obj.state = false;
  
  int slen = strlen(s);
  obj_needs(slen + 1);
  
  strcpy(obj.op, s);
  obj.op += slen;
  obj.ip += slen;
  return obj.state = true;
}


// skip if ip = s
bool skip(const char* s) {
  if(!is(s))
    return obj.state;
  int slen = strlen(s);
  obj.ip += slen;
  return obj.state = true;
}


bool copy_nchars(int n) {
  if(!is_nchars(n))
    return obj.state;
  
  obj_needs(n);
  while(n--)
    *obj.op++ = *obj.ip++;
  *obj.op = 0;
  return obj.state = true;
}

// true if input has n or more chars left
bool is_nchars(int n) {
  return obj.state = ((obj.ip + n) > (obj.iblimit)) ? false : true;
}

bool skip_nchars(int n)
{
  obj.ip += n;
  if(obj.ip < (obj.iblimit))
    return obj.state = true;
  return obj.state = false;
}

/// convenience functions for single chars
bool is_achar()
{
  // rarely used, not optimized for speed here.
  return is_nchars(1); 
}

bool skip_achar()
{
  if(obj.ip >= obj.iblimit)
    return obj.state = false;
  obj.ip++;
  return obj.state = true;
}

bool copy_achar()
{
  if(obj.ip >= obj.iblimit)
    return obj.state = false;
  
  if(obj.op >= obj.oblimit)
    obj_needs(1);
  
  *obj.op++ = *obj.ip++;
  return obj.state;
}


// whitespaces
bool is_ws() {
  if(!*obj.ip || !strchr(" \t\r", *obj.ip))
    return obj.state = false;
  return obj.state = true;
}

bool skip_ws() {
  if(!*obj.ip || !strchr(" \t\r", *obj.ip))
    return obj.state = false;
  obj.ip++;
  return obj.state = true;
}

bool copy_ws() {
  if(!is_ws())
    return obj.state;
  if(obj.op >= obj.oblimit)
    obj_needs(1);
  *obj.op++ = *obj.ip++;
  return obj.state = true;
}

// text to end of line, not including newline
bool is_linetext(){return not_eoi();} 

bool skip_linetext()
{
  if(is_eoi())
    return obj.state = false;
  // do all at once
  const char* eol = strchr(obj.ip, '\n');
  if(!eol)
    eol = obj.ip + strlen(obj.ip);
  skip_nchars(eol - obj.ip);
  return obj.state; // if nchars
}

bool copy_linetext()
{
  if(is_eoi())
    return obj.state = false;
  // do all at once
    const char* eol = strchr(obj.ip, '\n');
    if(!eol)
      eol = obj.ip + strlen(obj.ip);
    copy_nchars(eol - obj.ip);
    return obj.state; // if nchars
}

bool is_newline() { return is("\n"); }
bool skip_newline() { return skip("\n"); }
bool copy_newline() { return copy("\n"); }

// mid-level functions to handle delimited text (ignores escaped 
// chars, such as single and double quotes within a string), up 
// to but not including delimiter.  Delimiter can be any string.

// set's pointer to string in input or null if none, returns true if found
// scans forward
bool obj_find_ip(const char** ret, const char* s) {
  *ret = false;
  int slen = strlen(s);
//  const char* end = obj.ip + obj.inbuflen - (slen - 1);
  const char* end = obj.iblimit + 1 - slen;
  int first_chr = *s;
  
  const char* p = obj.ip;
  for( ; p < end; p++) {
    if((*p == first_chr) && memcmp((void*)p, (void*)s, slen) == 0) {
      if(p == obj.inbuf) {
        *ret = p;
        return obj.state = true;
      }
      else if(p[-1] != '\\') {
        *ret = p;
        return obj.state = true;
      }
    }
  }
  return obj.state = false;
}


bool copy_to(const char* delim) {
  const char* to;
  obj_find_ip(&to, delim);
  if(!obj.state) return false;

  int slen = to - obj.ip;
  obj_needs(slen);
  memcpy(obj.op, obj.ip, slen);
  obj.op += slen;
  *obj.op = 0; // terminate string
  obj.ip = to;
  return obj.state = true;
}

bool copy_past(const char* delim)
{
  if(copy_to(delim))
     copy(delim);
  return obj.state;
}

// true if exists in input
bool is_to(const char* delim) { 
  const char* dummy;
  if(obj_find_ip(&dummy, delim))
    return obj.state = true;
  return obj.state = false;
}

bool skip_to(const char* delim) {
  const char* to;
  obj_find_ip(&to, delim);
  if(!obj.state) return false;
  obj.ip = to;
  return obj.state; // true
}

bool skip_past(const char* delim) {
  if(skip_to(delim))
    obj.ip += strlen(delim);
  return obj.state; // true
}


//////////////////////////////////////////////////////////////

char* obj_get_result()
{
  *obj.op = 0;
  return obj.outbuf;
}


// returns the current offset into the output buffer
// ususally used to mark the beginning of text that
// may be of interest.
int obj_get_op()
{
  return obj.op - obj.outbuf;
}

int obj_get_ip()
{
  return obj.ip - obj.inbuf;
}

// for undoing a mark, resetting op to an offset
void obj_set_op(int offs)
{
  obj.op = obj.outbuf + offs;
}

void obj_set_ip(int offs)
{
  obj.ip = obj.inbuf + offs;
}

// used by OPT macro to make an evaluation optional.
bool obj_set_state(bool tf)
{
  return obj.state = tf;
}

/////////////////////////////////////////////////////////////
// partial set of parser funcs to deal with legal C names.
bool is_digit()      // sets state, same as isdigit(*obj.ip)
{
  return obj.state = isdigit(*obj.ip);
}

bool copy_cname()    // first is not digit all are cname chars.
{
  if(is_digit())
    return obj.state = false;
  int n = 0;
  for(; is_cname_char(); n++)
    copy_achar();
  return obj.state = (n>0);
}

bool is_cname_char() // after first char
{
  return obj.state = 
    isalnum(*obj.ip) || (*obj.ip == '_');
}

bool skip_cname()
{
  bool ret;
  int mark = obj_get_op();
  ret = copy_cname();
  obj_set_op(mark);
  return obj.state = ret;
}

// only needs to check first char to qualify as a cname
bool is_cname()
{
  return obj.state = !isdigit(*obj.ip) && is_cname_char();
}

/// See push_str(int offset) 
// copies accumulated output string from 'mark' and loads the 
// return buffer with a max of bufmax bytes.  Returns false only 
// if the buffer is too small, otherwise always true.
bool obj_get_opstr(char* retbuf, int bufmax, int mark)
{
  *obj.op = 0;              // terminate output buffer
  *retbuf = 0;              // null string at return buffer
  obj_set_op(mark);      // repoint op to mark
  char* text = obj.op;      // just a name
  int len = strlen(text);   // check the length
  if(len >= bufmax)
    return obj.state = false;
  strcpy(retbuf, text);     // copy the string
  *obj.op = 0;              // clear
  return obj.state = true;  // continue...
}

const char* obj_get_error()
{
  char* buf = tmpstr(strlen(obj.errormsg) + 512, 0);
  char* p = buf;
  char line[256];
  int offset;
  int line_no;
  p += sprintf(p, "Parse error: %s\n", obj.errormsg);
  obj_get_ipline(&line_no, &offset, line, 256);
  p += sprintf(p, "line: %d\n", line_no);
  p += sprintf(p, "%s\n", line);
  for(int i = 0; i <= offset; i++)
    p += sprintf(p, "-");
  p += sprintf(p, "^\n");
  return buf;
}

