// parser2.h - made easier to use, adds stacks and xt dictionary 

/* State machines might seem sexy at first glance but I can think of one (flex) 
 * that appears to have gonorhea.  And bison's not much better.
 * 
 * The Computer Mad Science Team
 */

#ifndef new_parser_h
#define new_parser_h

#ifndef PARSER2_NO_WARN
#warning  Use: "#define obj (*g_parser_obj)" in your source file(s) for the 'obj' name
#warning  Use: "#define PARSER2_NO_WARN" to silence this notice.
#endif // PARSER2_NO_WARN

/*

Gotcha's to look out for and some debugging tips:

 1. When printing obj_get_result() or obj_get_error() make sure you use the 
    "%s" format string or any "%s" in the text may trigger a seg fault.

 2. The best way to watch obj.op in a debugger is to set obj.op_watch' = obj.op
    in your main parser loop.  if you increment obj.op_ctr on each loop you can 
    also get a good idea where bugs are creeping in. (See also "long parses" 
    below.)

 3. If you need to set an error flag and return immediately consider setting
    up a jmp_buf (setjmp.h) and doing your parse like so:
      if(!setjmp(jmpbuf) && parse()) { <do success> }else{ <do failure> }.
    to allow jumping directly to the <do failure> routines.
    
 4. In long parses where errors occur WAY into the input buffer, you can 
    bump the obj.ctr so it will will tell you when you need to start watching 
    closely for the bug.  You can set your debugger to ignore the next <N> 
    hits, which will get you right to the action very quickly once you know 
    what <N> is.

 5. Multiple passes on a buffer can help isolate missteps in a parsing job.  While 
    this will potentially change the code needed in subsequent passes, that 
    would be the case anyway when done in a single pass.  Additionally, when 
    done in multiple passes special cases can be 'encoded' for ease of recognition 
    later.  For example if "12" is generated at some point as a word in a 
    specific context and it would be a number other contexts, you could encode 
    the word '12' (for example '\\wOrD[12]') so that it's never confused for a 
    number and the encodings can be removed in the last pass.

 */


// if using straight C we might need a few defs
#include <stdbool.h>  // bool, true, false
#include <stdlib.h>   // ulong, uint, etc.
#include <ctype.h>    // char tests

// the new stack and dictionary stuff 1K and 256K (by dword field sizes)
#define PARSER_STACK_SIZE 256
#define RETURN_STACK_SIZE 256
#define PARSER_DICT_SIZE (1024 * 64)

// executable tokens
typedef void(*xt_type)();

// typical parser func
typedef bool(*parser_func)();

typedef struct _PARSER_OBJ {
  // main watch points
  bool state;           /// current match state
  char* op_watch;       /// set every parser loop to watch op
  int op_ctr;           /// increment every loop (or???) to identify 'when' things happen
  int nest_level;       /// additional identification for debugging.

  // most interesting input vars
  const char* ip;       /// read-only input pointer
  const char* inbuf;    /// read-only input buffer
  const char* iblimit;  /// end of input buffer
  
  // general input and output info
  char* op;             /// output pointer
  char* outbuf;         /// output buffer base
  char* oblimit;        /// threshold for output buffer reallocation
  int outbuflen;        /// output buffer length
  int inbuflen;         /// input buffer length
  bool dynalloc;        /// input buffer can be freed
  _PARSER_OBJ* prev;    /// for nested parsers

  // extras
  bool errorflag;   /// error flag
  char* errormsg;   /// msg
  char** pstack;    /// stack space for strings/tokens
  int pstack_size;  /// PARSER_STACK_SIZE; // by char* fields
  int sp;           /// parameter stack pointer
  xt_type* dict;    /// executable dictionary, dword aligned
  int dict_size;    /// = PARSER_DICT_SIZE;  // by long fields
  int dp;           /// dictionary pointer/program counter
  xt_type* rstack;  /// return stack space for call-return
  int rstack_size;  /// RETURN_STACK_SIZE; // by char* fields
  int rp;           /// return stack pointer
}PARSER_OBJ;

typedef struct {
  const char* ip;   /// saved input pointer
  int svop;         /// saved output pointer offset
  int sp;           /// saved stack pointer
  int dp;           /// saved dictionary pointer
  int rp;           /// saved return pointer
}PPTRS_OBJ;

/** To be defined in the main file OR files creating new parser elements:

  #define obj (*g_parser_obj)

This is trick to allow dotted field addressing through the global pointer,
e,g, 'obj.fieldName' instead of 'g_parser_obj->fieldName'.  This is also a 
convenient watch point in a debugger for working out parsing algorithms.
*/

extern PARSER_OBJ* g_parser_obj;

/// Creates a new parser object and sets it as current, storing 
/// previous internally. The parsers must be deleted in the opposite
/// order in which they where created.  Set is_dyn = true if the 
/// input buffer was dynamically allocated and you want it freed 
/// automatically when the parser object is deleted.
PARSER_OBJ* parser_new(const char* input, int inlen, bool is_dyn);

/// frees memory including This, sets the global 
/// object to prev-ious and also returns it for 
/// convenience if you want it for something.
PARSER_OBJ* parser_delete(PARSER_OBJ* This);

/// restores a previous global object pointer
PARSER_OBJ* parser_set_current(PARSER_OBJ* This);

/// user supplies pointers to variables and a line buffer/len. The 
/// function sets the variables and fills the buffer with the 
/// terminated text of the line ip is on.  Pointers set to null will 
/// be ignored.
void obj_get_ipline(int* line_no, int* c_offs, char* linebuf, int bufmax);

/// internal function for SAVE_PTRS()
PPTRS_OBJ* pvt_parser_save_ptrs(PPTRS_OBJ* p);

/// internal function for RESTORE_PTRS() returns true flag but 
/// does not set the parser state var
bool pvt_parser_restore_ptrs(PPTRS_OBJ* p);

/// creates a save struct in current scope and saves ip/op pointers
#define SAVE_PTRS() \
  PPTRS_OBJ pvt_ptrs; pvt_ptrs = *pvt_parser_save_ptrs(&pvt_ptrs)

/// uses a save struct in current scope to restore ip/op pointers
#define RESTORE_PTRS() \
  pvt_parser_restore_ptrs(&pvt_ptrs)

/// built in formatted error string string (fixed width text best)
const char* obj_get_error();


////////////////////////////////////////////////////////
// built-in utilities acting on ip and/or op - prefix 'obj'
// is -- test IF all match (changes: none)
// copy -- test and copy to op IF all match (changes; ip, op)
// skip -- test and skip IF all match (changes: ip)

/// ip is string s, else false
bool is(const char* s);

/// copy s if at ip, else false
bool copy(const char* s);

/// skip s if at ip, else false
bool skip(const char* s);

/// true if there are n chars or more left to go
bool is_nchars(int n);

/// copy n chars from ip to op
bool copy_nchars(int n);

/// skip n chars or to end of buffer if n > what's left at ip.
bool skip_nchars(int n);

/// (safe) convenience functions for single chars
bool is_achar();
bool copy_achar();
bool skip_achar();

  
//////////////////////////////////////////////////////
// special functions, always return true

/// Dynamic reallocation check, always returns true.  Allocates
/// bytes as needed for output.
bool obj_needs(int numBytes);

/// put text in output buffer, always returns true
bool put(const char* text);

/// putfmt is like sprintf(...). It puts formatted text
/// at the output buffer and advances op.
bool putfmt(const char* fmt, ...);

///////////////////////////////////////////////////////////////
// built-in std parser funcs (no prefix)
// type bool func(void) for possible use by name in grammars)

/// true if end of input
bool is_eoi();   

/// true if not end of input
bool not_eoi();  

//////////////////////////////////////////////////////////////

/// is whitespace at ip, else false
bool is_ws();

/// copy whitespace at ip, else false
bool copy_ws();

/// skip whitespace at ip, else false
bool skip_ws();

//////////////////////////////////////////////////////////////

/// is text to end of line, not including newline, alias for not_eoi()
bool is_linetext();

/// copy text to end of line, not including newline
bool copy_linetext();

/// skip text to end of line, not including newline
bool skip_linetext();

//////////////////////////////////////////////////////////////

/// is a newline at ip, else false
bool is_newline();

/// copy a newline at ip, else false
bool copy_newline();

/// skip a newline at ip, else false
bool skip_newline();

//////////////////////////////////////////////////////////////

// functions to handle delimited text (handles escaped chars, 
// such as single and double quotes within a string), up to 
// but not including delimiter.  Delimiter can be any string.

/// delim exists in ip, else false
bool is_to(const char* delim);

/// copy to delim in ip, else false
bool copy_to(const char* delim);

/// skip to delim in ip, else false
bool skip_to(const char* delim);

/// copy past delim in ip, else false
bool copy_past(const char* delim);

/// skip past delim in ip, else false
bool skip_past(const char* delim);

// convenience funcs

/// set's pointer to string in input or null if none, returns true if found
/// \Note: scans forward
bool obj_find_ip(const char** ret, const char* s);

/// The safe way to return obj.obuf, makes sure the string is
/// terminated.  returns a pointer to obj.obuf, which must be 
/// copied before the the object mem is freed.
char* obj_get_result();

// get offset for current op
int obj_get_op();

// get offset for current ip
int obj_get_ip();

// set op to offset
void obj_set_op(int offs);

// set ip to offset
void obj_set_ip(int offs);

char* obj_op_str(int mark);


///////////////////////////////////////////////
// inlining optional parse elements

/// sometimes you want to set the global state to true, 
/// sometimes false for end of input tests
bool obj_set_state(bool tf);

// Sets state var = true and restores pointers undoing all 
// changes in the parser object.  This combines the OPT 
// macro and RESTORE_PTRS() so that it can be inlined as 
// a simple boolean function.
#define OPT_RESTORE_PTRS() || (obj.state = RESTORE_PTRS())

// See also OPT_RESTORE_PTRS()
#define OPT || obj_set_state(true)
// Usage example: ( copy(".") OPT )
// Must be used in a parenthetic parse evaluation.
// Will evaluate as true whether or not the operation is done, i.e., 
// makes it optional.  Must NOT be used in a 'while( blah OPT )' kind 
// of evaluation or it will loop endlessly or until it crashes.
 
// remove all trailing blanks at op
bool obj_trim_trailing();

/////////////////////////////////////////////////////////////
// partial set of parser funcs to deal with legal C names.
bool is_digit();      // sets state, same as isdigit(*obj.ip)
bool copy_cname();    // first is not digit all are cname chars.
bool is_cname_char(); // after first char
bool skip_cname();
bool is_cname();

/// copies accumulated output string from 'mark' and loads the 
/// return buffer with a max of bufmax bytes.  Returns false only 
/// if the buffer is too small, otherwise always true.
bool get_opstr(char* retbuf, int bufmax, int mark);

/// sets errormsg and errorflag
bool obj_set_errormsg(const char* fmt, ...);




// These will be implemented if we create a grammar for the parser2
// functions that looks somewhat like a bison file, but is more versatile.
#define BEGIN_INLINE not_implemented
#define END_INLINE not_implemented

#endif // new_parser_h

